@@ -6,6 +6,7 @@
 
 #include <stdio.h>
 #include <stdlib.h>
+#include "sanitizer_common/print_address.h"
 
 void **pp;
 
@@ -18,7 +19,7 @@ void *PutPointerOnStaleStack(void *p) {
   void *locals[2048];
   locals[0] = p;
   pp = &locals[0];
-  fprintf(stderr, "Test alloc: %p.\n", locals[0]);
+  print_address("Test alloc: ", 1, locals[0]);
   return 0;
 }
 
@@ -33,11 +34,11 @@ int main() {
 __attribute__((destructor))
 __attribute__((no_sanitize_address))
 void ConfirmPointerHasSurvived() {
-  fprintf(stderr, "Value after LSan: %p.\n", *pp);
+  print_address("Value after LSan: ", 1, *pp);
 }
-// CHECK: Test alloc: [[ADDR:.*]].
-// CHECK-sanity: Test alloc: [[ADDR:.*]].
+// CHECK: Test alloc: [[ADDR:0x[0-9,a-f]+]]
+// CHECK-sanity: Test alloc: [[ADDR:0x[0-9,a-f]+]]
 // CHECK: LeakSanitizer: detected memory leaks
 // CHECK: [[ADDR]] (1337 bytes)
 // CHECK: SUMMARY: {{(Leak|Address)}}Sanitizer:
-// CHECK-sanity: Value after LSan: [[ADDR]].
+// CHECK-sanity: Value after LSan: [[ADDR]]
