@@ -17,6 +17,10 @@ void callback(void *context, size_t i) {
 int main(int argc, const char *argv[]) {
   barrier_init(&barrier, 2);
   fprintf(stderr, "start\n");
+
+  // Warm up GCD (workaround for macOS Sierra where dispatch_apply might run single-threaded).
+  dispatch_sync(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^{ });
+
   dispatch_queue_t q = dispatch_queue_create("my.queue", DISPATCH_QUEUE_CONCURRENT);
 
   global = 42;
